@@ -2,8 +2,9 @@ import pandas as pd
 import numpy as np
 from pyg_encoders._parquet import pd_to_parquet, pd_read_parquet
 from pyg_encoders._encode import encode, decode
-from pyg_base import is_pd, is_dict, is_series, is_arr, is_date, dt2str, tree_items, dictable, try_value
+from pyg_base import is_pd, is_dict, is_series, is_arr, is_date, dt2str, tree_items, dictable, try_value, dt
 from pyg_npy import pd_to_npy, np_save, pd_read_npy, mkdir
+from pyg_base import Bi, bi_merge, is_bi, bi_read, try_none
 from functools import partial
 import pickle
 
@@ -79,11 +80,14 @@ def root_path_check(path):
         raise ValueError('The document did not contain enough keys to determine the path %s'%path)
     return path
 
-def pd_to_csv(value, path):
+def pd_to_csv(value, path, asof = None):
     """
     A small utility to write both pd.Series and pd.DataFrame to csv files
-    """
-    assert is_pd(value), 'cannot save non-pd'
+    """    
+    if '@' in path:
+        path, asof = path.split('@')    
+    if asof is not None:
+        value = Bi(value, asof)
     if is_series(value):
         value.index.name = _series
     if value.index.name is None:
@@ -91,12 +95,58 @@ def pd_to_csv(value, path):
     if path[-4:].lower()!=_csv:
         path = path + _csv
     mkdir(path)
+    if is_bi(value):
+        old = try_none(pd_read_csv)(path)
+        value = bi_merge(old, value)
     value.to_csv(path)
     return path
 
 
-def pickle_dump(value, path):
+def _pickle_raw(path):
+    try:
+        with open(path) as f:
+            df = pickle.load(f)
+    except Exception: #pandas read_pickle sometimes work when pickle.load fails
+        df = pd.read_pickle(path)
+    return df
+
+
+def pickle_dump(value, path, asof = None, existing_data = 'shift'):
+    """
+    saves a value as a pickle file
+    
+    Parameters
+    ----------
+    value: dict/dataframe
+        value to be saved
+    
+    asof:
+        set to a value if you want value to be converted to a bitemporal dataframe
+
+    existing_data:
+        if value is bitemporal, we need a policy how to handle existing data.
+        Further, existing data can be non bitemporal too. 
+            
+        'overwrite': overwrite existing
+        False / None: ignore non-bitemporal data, bi_merge if bitemporal 
+        other: apply bitemporal conversion to non-bitemporal data and then bi_merge
+        
+    """
+    if '@' in path:
+        path, asof = path.split('@')
     mkdir(path)
+    if asof is not None:
+        value = Bi(value, asof)
+    if is_bi(value):
+        if existing_data in ('ignore', 'overwrite'):
+            pass
+        else:            
+            old  = try_none(_pickle_raw)(path)
+            if old is not None:
+                if not is_bi(old) and existing_data:
+                    old = Bi(old, existing_data)
+                if is_bi(old):            
+                    value = bi_merge(old, value)
     if hasattr(value, 'to_pickle'):
         value.to_pickle(path) # use object specific implementation if available
     else:
@@ -104,19 +154,21 @@ def pickle_dump(value, path):
             pickle.dump(value, f)
     return path
 
-def pickle_load(path):
-    try:
-        with open(path) as f:
-            df = pickle.load(f)
-        return df
-    except Exception: #pandas read_pickle sometimes work when pickle.load fails
-        return pd.read_pickle(path) 
 
-def pd_read_csv(path):
+def pickle_load(path, asof = None, what = 'last'):
+    df = _pickle_raw(path)
+    if asof is not None:
+        df = bi_read(df, asof, what)
+    return df
+            
+
+def pd_read_csv(path, asof = None, what = 'last'):
     """
     A small utility to read both pd.Series and pd.DataFrame from csv files
     """
     res = pd.read_csv(path)
+    if asof is not None:
+        res = bi_read(res, asof, what)
     if res.columns[0] == _series and res.shape[1] == 2:
         res = pd.Series(res[res.columns[1]], res[_series].values)
         return res
@@ -152,37 +204,45 @@ _np_load = encode(np.load)
 _dictable_decode = encode(dictable_decode)
 
 
-def pickle_encode(value, path):
+def pickle_encode(value, path, asof = None):
     """
     encodes a single DataFrame or a document containing dataframes into a an abject of multiple pickled files that can be decoded
     """
+    if '@' in path:
+        path, asof = path.split('@')
     if path.endswith(_pickle):
         path = path[:-len(_pickle)]
     if path.endswith('/'):
         path = path[:-1]
     if is_pd(value):
         path = root_path_check(path)
-        return dict(_obj = _pickle_load, path = pickle_dump(value, path if path.endswith(_pickle) else path + _pickle))
+        path = path if path.endswith(_pickle) else path + _pickle
+        path = pickle_dump(value, path = path, asof = asof)
+        if asof is None:
+            return dict(_obj = _pickle_load, path = path)
+        else:
+            return dict(_obj = _pickle_load, path = path, asof = dt()) 
     elif is_arr(value):
         path = root_path_check(path)
         mkdir(path + _npy)
         np.save(path + _npy, value)
         return dict(_obj = _np_load, file = path + _npy)        
     elif is_dict(value):
-        res = type(value)(**{k : pickle_encode(v, '%s/%s'%(path,k)) for k, v in value.items()})
+        res = type(value)(**{k : pickle_encode(v, '%s/%s'%(path,k), asof) for k, v in value.items()})
         if isinstance(value, dictable):
             return dict(_obj = _dictable_decode,
-                        df = dict(_obj = _pickle_load, path = pickle_dump(res, path if path.endswith(_dictable) else path + _dictable)))
+                        df = dict(_obj = _pickle_load, 
+                                  path = pickle_dump(res, path if path.endswith(_dictable) else path + _dictable)))
         return res
     elif isinstance(value, (list, tuple)):
-        return type(value)([pickle_encode(v, '%s/%i'%(path,i)) for i, v in enumerate(value)])
+        return type(value)([pickle_encode(v, '%s/%i'%(path,i), asof) for i, v in enumerate(value)])
     else:
         return value
 
 pd_to_parquet_twice = try_value(pd_to_parquet, repeat = 2, sleep = 1, return_value = False)
 
 
-def parquet_encode(value, path, compression = 'GZIP'):
+def parquet_encode(value, path, compression = 'GZIP', asof = None):
     """
     encodes a single DataFrame or a document containing dataframes into a an abject that can be decoded
 
@@ -198,27 +258,34 @@ def parquet_encode(value, path, compression = 'GZIP'):
     >>> assert eq(decoded, value)
 
     """
+    if '@' in path:
+        path, asof = path.split('@')
     if path.endswith(_parquet):
         path = path[:-len(_parquet)]
     if path.endswith('/'):
         path = path[:-1]
     if is_pd(value):
         path = root_path_check(path)
-        return dict(_obj = _pd_read_parquet, path = pd_to_parquet_twice(value, path + _parquet))
+        path = pd_to_parquet_twice(value, path + _parquet, asof = asof)
+        if asof is None:
+            return dict(_obj = _pd_read_parquet, path = path)
+        else:
+            return dict(_obj = _pd_read_parquet, path = path, asof = dt())
     elif is_arr(value):
         path = root_path_check(path)
         mkdir(path + _npy)
         np.save(path + _npy, value)
         return dict(_obj = _np_load, file = path + _npy)        
     elif is_dict(value):
-        res = type(value)(**{k : parquet_encode(v, '%s/%s'%(path,k), compression) for k, v in value.items()})
+        res = type(value)(**{k : parquet_encode(v, '%s/%s'%(path,k), compression, asof = asof) for k, v in value.items()})
         if isinstance(value, dictable):
             df = pd.DataFrame(res)
             return dict(_obj = _dictable_decode,
-                        df = dict(_obj = _pd_read_parquet, path = pd_to_parquet_twice(df, path + _dictable)))
+                        df = dict(_obj = _pd_read_parquet, 
+                                  path = pd_to_parquet_twice(df, path + _dictable)))
         return res
     elif isinstance(value, (list, tuple)):
-        return type(value)([parquet_encode(v, '%s/%i'%(path,i), compression) for i, v in enumerate(value)])
+        return type(value)([parquet_encode(v, '%s/%i'%(path,i), compression, asof = asof) for i, v in enumerate(value)])
     else:
         return value
     
@@ -256,7 +323,7 @@ def npy_encode(value, path, append = False):
         return value
     
 
-def csv_encode(value, path):
+def csv_encode(value, path, asof = None):
     """
     encodes a single DataFrame or a document containing dataframes into a an abject that can be decoded while saving dataframes into csv
     
@@ -269,13 +336,19 @@ def csv_encode(value, path):
     >>> decoded = decode(encoded)
     >>> assert eq(decoded, value)
     """
+    if '@' in path:
+        path, asof = path.split('@')
     if path.endswith(_csv):
         path = path[:-len(_csv)]
     if path.endswith('/'):
         path = path[:-1]
     if is_pd(value):
         path = root_path_check(path)
-        return dict(_obj = _pd_read_csv, path = pd_to_csv(value, path))
+        path = pd_to_csv(value, path, asof = asof)
+        if asof is None:
+            return dict(_obj = _pd_read_csv, path = path)
+        else:
+            return dict(_obj = _pd_read_csv, path = path, asof = dt())
     elif is_dict(value):
         res = type(value)(**{k : csv_encode(v, '%s/%s'%(path,k)) for k, v in value.items()})
         if isinstance(value, dictable):
@@ -289,6 +362,27 @@ def csv_encode(value, path):
         return value
 
 def cell_root(doc, root = None):
+    """
+    finds a root parameter within a document. 
+
+    Priorities are:
+        1) if 'root' is in doc.db, we return it
+        2) if 'root' is in doc.root, we return it
+        3) return the original root
+
+    Parameters:
+    ----------
+    doc : dict
+        A document, such as a cell.
+    root : str, optional
+        The generic root location to save the document in.
+
+    Returns:
+    -------
+    root : str
+        The root location to save the document in.
+
+    """
     if _root in doc:
         root  = doc[_root]
     if root is None and _db in doc and isinstance(doc[_db], partial):
@@ -299,7 +393,8 @@ def cell_root(doc, root = None):
             root = keywords[_writer]
     return root
 
-def npy_write(doc, root = None, append = True):
+
+def npy_write(doc, root = None, append = True, asof = None):
     """
     MongoDB is great for manipulating/searching dict keys/values. 
     However, the actual dataframes in each doc, we may want to save in a file system. 
@@ -329,7 +424,7 @@ def npy_write(doc, root = None, append = True):
 
 
 
-def pickle_write(doc, root = None):
+def pickle_write(doc, root = None, asof = None):
     """
     MongoDB is great for manipulating/searching dict keys/values. 
     However, the actual dataframes in each doc, we may want to save in a file system. 
@@ -355,10 +450,10 @@ def pickle_write(doc, root = None):
     if root is None:
         return doc
     path = root_path(doc, root)
-    return pickle_encode(doc, path)
+    return pickle_encode(doc, path, asof = asof)
 
 
-def parquet_write(doc, root = None):
+def parquet_write(doc, root = None, asof = None):
     """
     MongoDB is great for manipulating/searching dict keys/values. 
     However, the actual dataframes in each doc, we may want to save in a file system. 
@@ -384,9 +479,9 @@ def parquet_write(doc, root = None):
     if root is None:
         return doc
     path = root_path(doc, root)
-    return parquet_encode(doc, path)
+    return parquet_encode(doc, path, asof = asof)
 
-def csv_write(doc, root = None):
+def csv_write(doc, root = None, asof = None):
     """
     MongoDB is great for manipulating/searching dict keys/values. 
     However, the actual dataframes in each doc, we may want to save in a file system. 
@@ -405,6 +500,6 @@ def csv_write(doc, root = None):
     if root is None:
         return doc
     path = root_path(doc, root)
-    return csv_encode(doc, path)
+    return csv_encode(doc, path, asof = asof)
 
 
